# Generated by Django 5.1.6 on 2025-02-23 14:28

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='QueryExport',
            fields=[
                ('id', models.CharField(max_length=64, primary_key=True, serialize=False, unique=True)),
                ('state', models.CharField(max_length=16)),
                ('metadata', models.JSONField()),
                ('file', models.FileField(upload_to='')),
                ('format', models.CharField(max_length=16)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
