# Generated by Django 5.1.6 on 2025-02-23 16:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('exporter', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='queryexport',
            name='state',
            field=models.CharField(choices=[('scheduled', 'Scheduled'), ('processing', 'Processing'), ('success', 'Success'), ('fail', 'Fail')], max_length=16),
        ),
    ]
