# Generated by Django 5.1.6 on 2025-02-22 15:14

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FeatureFlag',
            fields=[
                ('id', models.CharField(max_length=50, primary_key=True, serialize=False, unique=True)),
                ('active', models.BooleanField(default=False)),
                ('value', models.IntegerField(default=0)),
                ('value_str', models.CharField(default='', max_length=100)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
